import os
import logging
import json
from typing import Union, Dict, Any, List, Set

import pandas as pd

from dgraphpandas.types import find_dgraph_type, default_dgraph_type
from dgraphpandas.config import get_from_config

logger = logging.getLogger(__name__)


def _strip_id(original: Set[str]) -> Set[str]:
    temp_columns = set()
    for i in original:
        if str.endswith(i, '_id'):
            temp_columns.add(i[:-3])
        else:
            temp_columns.add(i)

    return temp_columns


def create_schema(source_config: Union[str, Dict[str, Any]], output_dir='.', **kwargs) -> pd.DataFrame:
    '''
    Creates a Schema DataFrame from the given Configuration.

    Parameters:
        config: a file path or already materialized dictionary
    '''
    if not source_config:
        raise ValueError('source_config')

    if isinstance(source_config, str):
        logger.debug(f'Loading configuration from {source_config}')
        with open(source_config, 'r') as f:
            config: Dict[str, Any] = json.load(f)
    else:
        config: Dict[str, Any] = source_config

    all_frames: List[pd.DataFrame] = []
    files: Dict[str, Any] = config['files']
    strip_id_from_edge_names: bool = get_from_config('strip_id_from_edge_names', config, True, **(kwargs))
    console: bool = get_from_config('console', config, False, **(kwargs))
    export_csv: bool = get_from_config('export_csv', config, False, **(kwargs))
    export_csv_name: str = get_from_config('export_csv_name', config, 'schema.csv', **(kwargs))
    ensure_xid_predicate: bool = get_from_config('ensure_xid_predicate', config, False, **(kwargs))

    for file, file_config in files.items():
        logger.debug(f'Preprocessing schema for {file}')

        subject_fields: List[str] = get_from_config('subject_fields', file_config, None, **(kwargs))
        override_edge_name: Dict[str, Any] = get_from_config('override_edge_name', file_config, {}, **(kwargs))
        list_edges: List[str] = get_from_config('list_edges', file_config, [], **(kwargs))

        if 'subject_fields' not in file_config:
            raise ValueError(f'{file} does not have subject_fields.')

        logger.debug('Building columns and types list')
        columns = set()
        columns.update(set(subject_fields))
        dgraph_types: Dict[str, str] = {}
        edge_fields: List[str] = []

        if 'type_overrides' in file_config:
            type_overrides: List[str] = get_from_config('type_overrides', file_config, None, **(kwargs))
            columns.update(set(type_overrides))
            dgraph_types = find_dgraph_type(file_config['type_overrides'])
        else:
            logger.warning(f'{file} does not have type_overrides skipping schema generation')

        if 'edge_fields' in file_config:
            edge_fields: List[str] = get_from_config('edge_fields', file_config, [], **(kwargs))
            columns.update(set(edge_fields))

        if 'csv_edges' in file_config:
            csv_edges: List[str] = get_from_config('csv_edges', file_config, [], **(kwargs))
            columns.update(set(csv_edges))
            edge_fields.extend(csv_edges)

        if 'ignore_fields' in file_config:
            ignore_fields: List[str] = get_from_config('ignore_fields', file_config, [], **(kwargs))
            columns = columns.difference(set(ignore_fields))

        if 'override_edge_name' in file_config:
            override_edge_name: Dict[str, Any] = get_from_config('override_edge_name', file_config, {}, **(kwargs))
            for options in override_edge_name.values():
                columns.add(options['predicate'])
                edge_fields.append(options['predicate'])

        if strip_id_from_edge_names:
            columns = _strip_id(columns)
            edge_fields = _strip_id(edge_fields)
            list_edges = _strip_id(list_edges)

        if 'pre_rename' in file_config:
            pre_rename: Dict[str, str] = get_from_config('pre_rename', file_config, {}, **(kwargs))
            columns = map(lambda x: pre_rename.get(x, x), columns)

            renamed_dgraph_types: Dict[str, str] = {}
            for original_name, new_name in pre_rename.items():
                renamed_dgraph_types[new_name] = dgraph_types.get(original_name, default_dgraph_type)
                if original_name in list_edges:
                    list_edges.add(new_name)
                if original_name in edge_fields:
                    edge_fields.add(new_name)

                dgraph_types.update(renamed_dgraph_types)

        logger.debug('Building Schema DataFrame')

        frame = pd.DataFrame(columns=['column'])
        frame['column'] = list(columns)
        frame['type'] = frame['column'].map(dgraph_types)

        frame['type'] = frame['type'].fillna(default_dgraph_type)
        frame.loc[frame['column'].isin(edge_fields), 'type'] = 'uid'
        frame.loc[frame['column'].isin(list_edges), 'type'] = '[uid]'

        frame['table'] = file

        if 'options' in file_config:
            options: Dict[str, str] = get_from_config('options', file_config, {}, **(kwargs))
            options = {column: str.join(' ', options) for column, options in options.items() if options is not None}
            frame['options'] = frame['column'].map(options)
        else:
            frame['options'] = None

        all_frames.append(frame)

    if not all_frames:
        logger.warning('No frames were generated')
        return

    frame: pd.DataFrame = pd.concat(all_frames)

    logger.debug('Appending xid declaration')
    if ensure_xid_predicate:
        # Create a new DataFrame for the row to be added
        new_row = pd.DataFrame([{'column': 'xid', 'type': 'string', 'table': None, 'options': '@index(exact)'}])
        # Use pd.concat to add the new row
        frame = pd.concat([frame, new_row], ignore_index=True)

    frame.sort_values(by=['table', 'type'], inplace=True)
    frame.reset_index(inplace=True, drop=True)

    if console:
        print(frame)
    if export_csv:
        path = os.path.join(output_dir, export_csv_name)
        logger.info(f'Writing pre schema file to {path}')
        frame.to_csv(path, index=False)

    return frame



from unittest.mock import patch, Mock, mock_open

import pytest
from parameterized import parameterized
import pandas as pd

from dgraphpandas.writers.types import generate_types


@parameterized.expand([
    ('without_frame', None),
    ('without_column', pd.DataFrame(columns=['type', 'table', 'options'])),
    ('without_type', pd.DataFrame(columns=['column', 'table', 'options'])),
    ('without_table', pd.DataFrame(columns=['column', 'type', 'options'])),
    ('with_options', pd.DataFrame(columns=['column', 'type', 'table'])),
])
def test_invalid_parameters(name, frame: pd.DataFrame):
    '''
    Ensures when invalid parameters are given,
    then an error is raised
    '''
    with pytest.raises(ValueError, match='frame|column|type|options|table'):
        generate_types(frame)


def test_empty_valid_frame():
    '''
    Ensures when the frame is valid but empty,
    an empty list is returned
    '''
    frame = pd.DataFrame(columns=['column', 'type', 'table', 'options'])
    result = generate_types(frame)
    assert result == []


@parameterized.expand([
    ###
    (
        'single_table',
        pd.DataFrame(
            columns=['column', 'type', 'table', 'options'],
            data=[
                ('id', 'string', 'customer', '@index(exact)'),
                ('age', 'int', 'customer', None),
                ('dob', 'datetime', 'customer', None),
            ]),
        ["type customer { \nid\nage\ndob\n }\n"]
    ),
    ###
    (
        'multiple_tables',
        pd.DataFrame(
            columns=['column', 'type', 'table', 'options'],
            data=[
                ('id', 'string', 'customer', '@index(exact)'),
                ('age', 'int', 'customer', None),
                ('dob', 'datetime', 'customer', None),
                ('id', 'string', 'order', '@index(hash)'),
                ('value', 'double', 'order', None),
                ('product_id', 'string', 'product', '@index(hash)'),
                ('name', 'double', 'product', None),
            ]),
        ["type customer { \nid\nage\ndob\n }\n", "type order { \nid\nvalue\n }\n", "type product { \nproduct_id\nname\n }\n"]
    ),
    ###
    (
        'with_reverse',
        pd.DataFrame(
            columns=['column', 'type', 'table', 'options'],
            data=[
                ('id', 'string', 'customer', '@index(exact)'),
                ('age', 'int', 'customer', None),
                ('order', '[uid]', 'customer', '@reverse'),
                ('gender', '[uid]', 'customer', '@reverse'),
                ('id', 'string', 'order', '@index(hash)'),
                ('value', 'double', 'order', None),
                ('invoice', 'uid', 'order', '@reverse'),
            ]),
        ["type customer { \nid\nage\n<~order>\n<~gender>\n }\n", "type order { \nid\nvalue\n<~invoice>\n }\n"]
    )
])
def test_types_generated(name, frame, expected_result):
    '''
    Ensures types are generated when a valid frame is passed
    '''
    assert generate_types(frame) == expected_result


@patch('builtins.open', callable=mock_open)
def test_types_export_schema(mock_open: Mock):
    '''
    Ensures when export schema is passed then
    the file is written
    '''
    frame = pd.DataFrame(
        columns=['column', 'type', 'table', 'options'],
        data=[
            ('id', 'string', 'customer', '@index(exact)'),
            ('age', 'int', 'customer', None),
            ('order', '[uid]', 'customer', '@reverse'),
            ('gender', '[uid]', 'customer', '@reverse'),
            ('id', 'string', 'order', '@index(hash)'),
            ('value', 'double', 'order', None),
            ('invoice', 'uid', 'order', '@reverse'),
        ])

    expected = ["type customer { \nid\nage\n<~order>\n<~gender>\n }\n", "type order { \nid\nvalue\n<~invoice>\n }\n"]

    result = generate_types(frame, export_schema=True)

    assert result == expected

    args, kwargs = mock_open.call_args_list[0]
    assert args == ('./types.txt', 'w')
    assert kwargs == {'encoding': 'utf-8'}


def test_console(capsys):
    '''
    Ensures when console is provided, then the type builder
    is written to stdout
    '''
    frame = pd.DataFrame(
        columns=['column', 'type', 'table', 'options'],
        data=[
            ('id', 'string', 'customer', '@index(exact)'),
            ('age', 'int', 'customer', None),
            ('order', '[uid]', 'customer', '@reverse'),
            ('gender', '[uid]', 'customer', '@reverse'),
            ('id', 'string', 'order', '@index(hash)'),
            ('value', 'double', 'order', None),
            ('invoice', 'uid', 'order', '@reverse'),
        ])

    expected = ["type customer { \nid\nage\n<~order>\n<~gender>\n }\n", "type order { \nid\nvalue\n<~invoice>\n }\n"]

    result = generate_types(frame, console=True)

    assert result == expected

    out, err = capsys.readouterr()
    for ex in expected:
        assert ex in out



from unittest.mock import patch, Mock, mock_open

import pytest
import pandas as pd
from pandas.testing import assert_frame_equal
from parameterized import parameterized

from dgraphpandas.strategies.schema import create_schema


def test_create_schema_none_source_config():
    '''
    Ensures when the source config is None, then an error is raised
    '''
    with pytest.raises(ValueError):
        create_schema(None)


@patch('dgraphpandas.strategies.schema.json.load')
@patch('builtins.open', callable=mock_open)
def test_create_schema_no_files(mock_file: Mock, mock_json_loads: Mock):
    '''
    Ensures when no files is defined in config then an error is raised
    '''
    mock_json_loads.return_value = {"some_other_entry": {}}
    with pytest.raises(KeyError, match='files'):
        create_schema('my_config.json')

    mock_file.assert_called()
    mock_json_loads.assert_called()

    args, kwargs = mock_file.call_args_list[0]
    assert args == ('my_config.json', 'r')
    assert kwargs == {}


@patch('dgraphpandas.strategies.schema.json.load')
@patch('builtins.open', callable=mock_open)
def test_create_schema_file_path_config(mock_file: Mock, mock_json_loads: Mock):
    '''
    Ensures when a file path is passed which has files attribute, is it loaded
    but nothing is returned
    '''
    mock_json_loads.return_value = {"files": {}}
    result = create_schema('my_config.json')

    assert result is None

    mock_file.assert_called()
    mock_json_loads.assert_called()

    args, kwargs = mock_file.call_args_list[0]
    assert args == ('my_config.json', 'r')
    assert kwargs == {}


@parameterized.expand([
    ###
    (
        'single_file',
        {
            'files': {
                "animal": {
                    'edge_fields': ['habitat']
                }
            }
        },
        'animal does not have subject_fields.'
    ),
    ###
    (
        'multiple_files_after_first_has_no_subject_fields',
        {
            'files': {
                "animal": {
                    'subject_fields': ['animal_id'],
                    'edge_fields': ['habitat']
                },
                "habitat": {
                    'edge_fields': ['animal_id']
                }
            }
        },
        'habitat does not have subject_fields.'
    ),
    ###
    (
        'multiple_files_first_has_no_subject_fields',
        {
            'files': {
                "habitat": {
                    'edge_fields': ['animal_id']
                },
                "animal": {
                    'subject_fields': ['animal_id'],
                    'edge_fields': ['habitat']
                }
            }
        },
        'habitat does not have subject_fields.'
    ),
])
def test_create_schema_missing_subject_fields(name, config, expected_error_message):
    '''
    Ensures when the schema is missing subject fields, then an error is raised
    '''
    with pytest.raises(ValueError, match=expected_error_message):
        create_schema(config)


@parameterized.expand([
    ###
    (
        'only_subject_fields',
        {
            'files': {
                "animal": {
                    'subject_fields': ['animal_id'],
                }
            }
        },
        pd.DataFrame(data={
            'column': ['animal'],
            'type': ['string'],
            'table': 'animal',
            'options': None
        })
    ),
    ###
    (
        'with_type_overrides',
        {
            'files': {
                "animal": {
                    'subject_fields': ['animal_id'],
                    'type_overrides': {
                        'legs': 'int',
                        'weight': 'float'
                    }
                }
            }
        },
        pd.DataFrame(data={
            'column': ['animal', 'legs', 'weight'],
            'type': ['string', 'int', 'float'],
            'table': ['animal']*3,
            'options': [None]*3
        })
    ),
    ###
    (
        'with_edge_fields',
        {
            'files': {
                "animal": {
                    'subject_fields': ['animal_id'],
                    'type_overrides': {
                        'legs': 'int',
                        'weight': 'float'
                    },
                    'edge_fields': ['habitat_id']
                }
            }
        },
        pd.DataFrame(data={
            'column': ['animal', 'legs', 'weight', 'habitat'],
            'type': ['string', 'int', 'float', 'uid'],
            'table': ['animal']*4,
            'options': [None]*4
        })
    ),
    ###
    (
        'with_csv_edges',
        {
            'files': {
                "animal": {
                    'subject_fields': ['animal_id'],
                    'type_overrides': {
                        'legs': 'int',
                        'weight': 'float'
                    },
                    'edge_fields': ['habitat_id'],
                    'csv_edges': ['group_id']
                }
            }
        },
        pd.DataFrame(data=[
            ('animal', 'string', 'animal', None),
            ('legs', 'int', 'animal', None),
            ('weight', 'float', 'animal', None),
            ('habitat', 'uid', 'animal', None),
            ('group', 'uid', 'animal', None),
        ], columns=['column', 'type', 'table', 'options'])
    ),
    ###
    (
        'with_ignore_fields',
        {
            'files': {
                "animal": {
                    'subject_fields': ['animal_id'],
                    'type_overrides': {
                        'legs': 'int',
                        'weight': 'float'
                    },
                    'edge_fields': ['habitat_id', 'group_id'],
                    'ignore_fields': ['group_id']
                }
            }
        },
        pd.DataFrame(data=[
            ('animal', 'string', 'animal', None),
            ('legs', 'int', 'animal', None),
            ('weight', 'float', 'animal', None),
            ('habitat', 'uid', 'animal', None),
        ], columns=['column', 'type', 'table', 'options'])
    ),
    ###
    (
        'with_override_edge_fields',
        {
            'files': {
                "animal": {
                    'subject_fields': ['animal_id'],
                    'type_overrides': {
                        'legs': 'int',
                        'weight': 'float'
                    },
                    'edge_fields': ['habitat_id'],
                    'override_edge_name': {
                        'habitat_id': {
                            'predicate': 'habo'
                        }
                    }
                }
            }
        },
        pd.DataFrame(data=[
            ('animal', 'string', 'animal', None),
            ('legs', 'int', 'animal', None),
            ('weight', 'float', 'animal', None),
            ('habo', 'uid', 'animal', None),
            ('habitat', 'uid', 'animal', None),
        ], columns=['column', 'type', 'table', 'options'])
    ),
    ###
    (
        'without_strip_id_from_edge_names',
        {
            'strip_id_from_edge_names': False,
            'files': {
                "animal": {
                    'subject_fields': ['animal_id'],
                    'type_overrides': {
                        'legs': 'int',
                        'weight': 'float'
                    },
                    'edge_fields': ['habitat_id']
                }
            }
        },
        pd.DataFrame(data=[
            ('animal_id', 'string', 'animal', None),
            ('legs', 'int', 'animal', None),
            ('weight', 'float', 'animal', None),
            ('habitat_id', 'uid', 'animal', None),
        ], columns=['column', 'type', 'table', 'options'])
    ),
    ###
    (
        'with_prerename',
        {
            'files': {
                "animal": {
                    'subject_fields': ['animal_id'],
                    'type_overrides': {
                        'legs': 'int',
                        'weight': 'float'
                    },
                    'edge_fields': ['habitat_id'],
                    'pre_rename': {
                        'legs': 'number_of_legs',
                        'habitat': 'habo'
                    }
                }
            }
        },
        pd.DataFrame(data=[
            ('animal', 'string', 'animal', None),
            ('number_of_legs', 'int', 'animal', None),
            ('weight', 'float', 'animal', None),
            ('habo', 'uid', 'animal', None),
        ], columns=['column', 'type', 'table', 'options'])
    ),
    ###
    (
        'with_prename_on_list_edge',
        {
            'files': {
                "animal": {
                    'subject_fields': ['animal_id'],
                    'type_overrides': {
                        'legs': 'int',
                        'weight': 'float'
                    },
                    'edge_fields': ['habitat_id'],
                    'pre_rename': {
                        'legs': 'number_of_legs',
                        'habitat': 'habo'
                    },
                    'list_edges': ['habitat_id']
                }
            }
        },
        pd.DataFrame(data=[
            ('animal', 'string', 'animal', None),
            ('number_of_legs', 'int', 'animal', None),
            ('weight', 'float', 'animal', None),
            ('habo', '[uid]', 'animal', None),
        ], columns=['column', 'type', 'table', 'options'])
    ),
    ###
    (
        'with_options',
        {
            'files': {
                "animal": {
                    'subject_fields': ['animal_id'],
                    'type_overrides': {
                        'legs': 'int',
                        'weight': 'float'
                    },
                    'edge_fields': ['habitat_id'],
                    'options': {
                        'legs': ['@count', '@index(exact)'],
                        'habitat': ['@reverse']
                    }
                }
            }
        },
        pd.DataFrame(data=[
            ('animal', 'string', 'animal', None),
            ('legs', 'int', 'animal', '@count @index(exact)'),
            ('weight', 'float', 'animal', None),
            ('habitat', 'uid', 'animal', '@reverse'),
        ], columns=['column', 'type', 'table', 'options'])
    ),
    ###
    (
        'with_multiple_files',
        {
            'files': {
                "animal": {
                    'subject_fields': ['animal_id'],
                    'type_overrides': {
                        'legs': 'int',
                        'weight': 'float'
                    },
                    'edge_fields': ['habitat_id'],
                    'options': {
                        'legs': ['@count', '@index(exact)'],
                        'habitat': ['@reverse']
                    }
                },
                'habitat': {
                    'subject_fields': ['habitat_id'],
                    'type_overrides': {
                        'color': 'string',
                        'sea_level': 'int'
                    },
                    'edge_fields': ['country_id'],
                    'options': {
                        'color': ['@index(hash)']
                    }
                }
            }
        },
        pd.DataFrame(data=[
            ('animal', 'string', 'animal', None),
            ('legs', 'int', 'animal', '@count @index(exact)'),
            ('weight', 'float', 'animal', None),
            ('habitat', 'uid', 'animal', '@reverse'),
            ('habitat', 'string', 'habitat', None),
            ('color', 'string', 'habitat', '@index(hash)'),
            ('country', 'uid', 'habitat', None),
            ('sea_level', 'int', 'habitat', None),
        ], columns=['column', 'type', 'table', 'options'])
    ),

    ###
    (
        'with_list_edges',
        {
            'files': {
                "animal": {
                    'subject_fields': ['animal_id'],
                    'type_overrides': {
                        'legs': 'int',
                        'weight': 'float'
                    },
                    'edge_fields': ['habitat_id', 'predator_id'],
                    'list_edges': ['predator_id']
                }
            }
        },
        pd.DataFrame(data=[
            ('animal', 'string', 'animal', None),
            ('legs', 'int', 'animal'),
            ('weight', 'float', 'animal', None),
            ('habitat', 'uid', 'animal'),
            ('predator', '[uid]', 'animal'),
        ], columns=['column', 'type', 'table', 'options'])
    ),

])
def test_create_schema(name, config, expected_frame):
    '''
    Ensures given the parameterized config, the output frame matches the set expected_frame
    '''
    frame = create_schema(config)

    frame = frame.sort_values(by=['table', 'column']).reset_index(drop=True)
    expected_frame = expected_frame.sort_values(by=['table', 'column']).reset_index(drop=True)

    try:
        assert_frame_equal(frame, expected_frame)
    except AssertionError:  # pragma: no cover
        print('Actual Frame \n', frame)  # pragma: no cover
        print('####')  # pragma: no cover
        print('Expected Frame \n', expected_frame)  # pragma: no cover
        raise  # pragma: no cover


@patch('dgraphpandas.strategies.schema.pd.DataFrame.to_csv')
def test_create_schema_export_csv(to_csv_mock: Mock):
    '''
    Ensures when export_csv is passed, then the schema frame
    is passed into to_csv
    '''
    config = {
        'files': {
            "animal": {
                'subject_fields': ['id'],
                'edge_fields': ['habitat']
            }
        }
    }

    result = create_schema(config, export_csv=True)
    assert to_csv_mock.called

    args, kwargs = to_csv_mock.call_args_list[0]
    assert ('./schema.csv',) == args
    assert {'index': False} == kwargs

    expected_frame = pd.DataFrame(
        columns=['column', 'type', 'table', 'options'],
        data=[
            ('id', 'string', 'animal', None),
            ('habitat', 'uid', 'animal', None),
        ])

    assert_frame_equal(expected_frame, result)


def test_create_schema_ensure_xid():
    '''
    Ensures when ensure_xid_predicate is passed,
    then an xid predicate is appended to the schema
    '''
    config = {
        'files': {
            "animal": {
                'subject_fields': ['id'],
                'edge_fields': ['habitat']
            }
        }
    }

    expected_frame = pd.DataFrame(
        columns=['column', 'type', 'table', 'options'],
        data=[
            ('id', 'string', 'animal', None),
            ('habitat', 'uid', 'animal', None),
            ('xid', 'string', None, '@index(exact)'),
        ])

    result = create_schema(config, ensure_xid_predicate=True)
    assert_frame_equal(expected_frame, result)


@patch('builtins.print')
def test_create_schema_console(mock_print):
    '''
    Ensures when console is passed, then the frame is written to stdout.
    '''
    config = {
        'files': {
            "animal": {
                'subject_fields': ['id'],
                'edge_fields': ['habitat']
            }
        }
    }
    expected_frame = pd.DataFrame(
        columns=['column', 'type', 'table', 'options'],
        data=[
            ('id', 'string', 'animal', None),
            ('habitat', 'uid', 'animal', None),
        ])

    result = create_schema(config, console=True)
    assert_frame_equal(expected_frame, result)

    assert mock_print.called
    args, kwargs = mock_print.call_args_list[0]
    assert_frame_equal(args[0], expected_frame)
    assert kwargs == {}



from unittest.mock import patch, Mock
from typing import List

import pytest
from parameterized import parameterized

from dgraphpandas import __version__
from dgraphpandas.__main__ import main


@patch('dgraphpandas.__main__.sys')
def test_no_arguments(argv_mock: Mock, capsys):
    '''
    Ensures when no arguments are passed, then usage is outputted
    '''
    with pytest.raises(SystemExit, match='2'):
        argv_mock.argv = ['script']
        main()

    assert 'usage:' in capsys.readouterr().err


@patch('dgraphpandas.__main__.sys')
def test_missing_required_arguments(argv_mock: Mock, capsys):
    '''
    Ensures when the required arguments are missing then usage is outputted
    '''
    with pytest.raises(SystemExit, match='2'):
        argv_mock.argv = [
            'script',
            '-x', 'upsert',
            '-f', 'my_file.csv',
            # '-c', 'my_config.json' # <- missing config
        ]
        main()

    assert 'usage:' in capsys.readouterr().err


@patch('dgraphpandas.__main__.logging')
@patch('dgraphpandas.__main__.to_rdf')
@patch('dgraphpandas.__main__.sys')
def test_bad_operation(
        argv_mock: Mock,
        to_rdf_mock: Mock,
        logger_mock: Mock,
        capsys):
    '''
    Ensures when the operation is unknown, then usage is raised
    '''
    argv_mock.argv = [
        'script',
        '-x', 'unknown',
        '-f', 'my_file.csv',
        '-c', 'my_config.json'
    ]

    with pytest.raises(SystemExit, match='2'):
        main()

    assert "invalid choice: 'unknown'" in capsys.readouterr().err


@patch('dgraphpandas.__main__.logging')
@patch('dgraphpandas.__main__.to_rdf')
@patch('dgraphpandas.__main__.sys')
def test_upsert_missing_config_file_key(
        argv_mock: Mock,
        to_rdf_mock: Mock,
        logger_mock: Mock,
        capsys):
    '''
    Ensures when config_file_key is not provided then an error is raised
    '''
    argv_mock.argv = [
        'script',
        '-x', 'upserts',
        '-f', 'my_file.csv',
        '-c', 'my_config.json'
    ]

    with pytest.raises(ValueError, match='config_file_key must be provided in upsert mode'):
        main()


@patch('dgraphpandas.__main__.logging')
@patch('dgraphpandas.__main__.to_rdf')
@patch('dgraphpandas.__main__.sys')
def test_upsert_missing_file(
        argv_mock: Mock,
        to_rdf_mock: Mock,
        logger_mock: Mock,
        capsys):
    '''
    Ensures when file is not provided then an error is raised
    '''
    argv_mock.argv = [
        'script',
        '-x', 'upserts',
        '-c', 'config.json',
        '-ck', 'my_key'
    ]

    with pytest.raises(ValueError, match='file must be provided in upsert mode'):
        main()


@patch('dgraphpandas.__main__.logging')
@patch('dgraphpandas.__main__.to_rdf')
@patch('dgraphpandas.__main__.sys')
def test_upsert(
        argv_mock: Mock,
        to_rdf_mock: Mock,
        logger_mock: Mock,
        capsys):
    '''
    Ensures when upsert is called with required parameters
    then to_rdf is called
    '''
    argv_mock.argv = [
        'script',
        '-x', 'upserts',
        '-c', 'config.json',
        '-ck', 'my_key',
        '-f', 'my_file'
    ]

    main()

    assert to_rdf_mock.called
    args, kwargs = to_rdf_mock.call_args_list[0]

    assert args == ('my_file', 'config.json', 'my_key', '.')
    assert kwargs == {
        'export_rdf': True,
        'add_dgraph_type_records': True,
        'drop_na_intrinsic_objects': True,
        'drop_na_edge_objects': True,
        'illegal_characters': ['%', '\\.', '\\s', '"', '\\n', '\\r\\n'],
        'illegal_characters_intrinsic_object': ['"', '\\n', '\\r\\n'],
        'chunk_size': 10000000
    }


@patch('dgraphpandas.__main__.logging')
@patch('dgraphpandas.__main__.generate_schema')
@patch('dgraphpandas.__main__.create_schema')
@patch('dgraphpandas.__main__.sys')
def test_schema(
        argv_mock: Mock,
        create_schema_mock: Mock,
        generate_schema_mock: Mock,
        logger_mock: Mock,
        capsys):
    '''
    Ensure when schema operation is called with the valid args
    then generate_schema is called
    '''
    argv_mock.argv = [
        'script',
        '-x', 'schema',
        '-c', 'config.json',
    ]

    create_schema_mock.return_value = 'fake_schema'

    main()

    assert create_schema_mock.called
    assert generate_schema_mock.called

    args, kwargs = create_schema_mock.call_args_list[0]

    assert args == ('config.json',)
    assert kwargs == {
        'add_dgraph_type_records': True,
        'drop_na_intrinsic_objects': True,
        'drop_na_edge_objects': True,
        'illegal_characters': ['%', '\\.', '\\s', '"', '\\n', '\\r\\n'],
        'illegal_characters_intrinsic_object': ['"', '\\n', '\\r\\n'],
        'chunk_size': 10000000,
        'ensure_xid_predicate': True
    }

    args, kwargs = generate_schema_mock.call_args_list[0]

    assert args == ('fake_schema',)
    assert kwargs == {
        'add_dgraph_type_records': True,
        'drop_na_intrinsic_objects': True,
        'drop_na_edge_objects': True,
        'illegal_characters': ['%', '\\.', '\\s', '"', '\\n', '\\r\\n'],
        'illegal_characters_intrinsic_object': ['"', '\\n', '\\r\\n'],
        'chunk_size': 10000000,
        'export_schema': True
    }


@parameterized.expand([
    ([], 'INFO'),
    (['--verbosity', 'DEBUG'], 'DEBUG'),
    (['--verbosity', 'INFO'], 'INFO'),
    (['--verbosity', 'WARNING'], 'WARNING'),
    (['--verbosity', 'ERROR'], 'ERROR'),
    (['--verbosity', 'NOTSET'], 'NOTSET'),
])
@patch('dgraphpandas.__main__.logging')
@patch('dgraphpandas.__main__.generate_schema')
@patch('dgraphpandas.__main__.create_schema')
@patch('dgraphpandas.__main__.sys')
def test_logging_verbosity_set(
        logging_options: List[str],
        expected_set_level: str,
        argv_mock: Mock,
        create_schema_mock: Mock,
        generate_schema_mock: Mock,
        logger_mock: Mock):
    '''
    Ensures when logging is set via cli then its passed
    to the logger
    '''
    argv_mock.argv = [
        'script',
        '-x', 'schema',
        '-c', 'config.json'
    ]
    argv_mock.argv.extend(logging_options)

    main()

    args, kwargs = logger_mock.basicConfig.call_args_list[0]

    assert not args
    assert kwargs == {'level': expected_set_level}


@patch('dgraphpandas.__main__.sys')
def test_version(
        argv_mock: Mock,
        capsys):
    '''
    Ensures when the version is requested then it is printed
    to stdout
    '''
    argv_mock.argv = ['script', '--version']

    with pytest.raises(SystemExit, match='0'):
        main()

    assert __version__ in capsys.readouterr().out


@patch('dgraphpandas.__main__.logging')
@patch('dgraphpandas.__main__.generate_types')
@patch('dgraphpandas.__main__.create_schema')
@patch('dgraphpandas.__main__.sys')
def test_types(
        argv_mock: Mock,
        create_schema_mock: Mock,
        generate_types_mock: Mock,
        logger_mock: Mock,
        capsys):
    '''
    Ensures when types is called, the underlying services
    are called
    '''
    argv_mock.argv = [
        'script',
        '-x', 'types',
        '-c', 'config.json',
    ]

    create_schema_mock.return_value = 'fake_schema'

    main()

    assert create_schema_mock.called
    assert generate_types_mock.called

    args, kwargs = create_schema_mock.call_args_list[0]

    assert args == ('config.json',)
    assert kwargs == {
        'add_dgraph_type_records': True,
        'drop_na_intrinsic_objects': True,
        'drop_na_edge_objects': True,
        'illegal_characters': ['%', '\\.', '\\s', '"', '\\n', '\\r\\n'],
        'illegal_characters_intrinsic_object': ['"', '\\n', '\\r\\n'],
        'chunk_size': 10000000,
        'ensure_xid_predicate': True
    }

    args, kwargs = generate_types_mock.call_args_list[0]

    assert args == ('fake_schema',)
    assert kwargs == {
        'add_dgraph_type_records': True,
        'drop_na_intrinsic_objects': True,
        'drop_na_edge_objects': True,
        'illegal_characters': ['%', '\\.', '\\s', '"', '\\n', '\\r\\n'],
        'illegal_characters_intrinsic_object': ['"', '\\n', '\\r\\n'],
        'chunk_size': 10000000,
        'export_schema': True
    }



import os
import logging
from typing import List

import pandas as pd

logger = logging.getLogger(__name__)


def generate_types(frame: pd.DataFrame, **kwargs) -> List[str]:
    '''
    Given the pre-processed DataFrame from the schema
    strategy, generate types.
    '''
    if frame is None:
        raise ValueError('frame')
    if 'column' not in frame:
        raise ValueError('column')
    if 'type' not in frame:
        raise ValueError('type')
    if 'table' not in frame:
        raise ValueError('table')
    if 'options' not in frame:
        raise ValueError('options')

    output_dir = kwargs.get('output_dir', '.')
    export_schema = kwargs.get('export_schema', False)
    export_file = kwargs.get('export_file', 'types.txt')
    console = kwargs.get('console', False)
    encoding = kwargs.get('encoding', 'utf-8')
    line_delimeter = kwargs.get('line_delimeter ', '\n')

    all_types: List[str] = []
    all_types_reverse: List[str] = []

    tables = frame.groupby(by=['table'])
    for name, current_frame in tables:
        logger.debug(f'Creating types for {name}')

        reverse_edge_mask = (~current_frame['options'].isnull()) & current_frame['options'].str.contains('@reverse')
        current_frame.loc[reverse_edge_mask, 'column'] = '<~' + current_frame['column'] + '>'

        type_builder = 'type ' + name
        type_builder += ' { '
        type_builder += line_delimeter
        type_builder += line_delimeter.join(current_frame['column'].unique().tolist())
        type_builder += line_delimeter
        type_builder += ' }'
        type_builder += line_delimeter

        # Split up types with reverse edges so we can gurantee they are applied after other types
        # This is required because if dgraph live encounters a reverse edge for a type defined later in the file
        # then dgraph live will fails.
        # NOTE: There might be a better solution here
        # and we could build a dependency tree based on the references
        # topological sort?
        # also this won't detect circular dependencies
        if current_frame.loc[reverse_edge_mask, 'column'].shape[0]:
            all_types_reverse.append(type_builder)
        else:
            all_types.append(type_builder)

        if console:
            print(type_builder)
            print(line_delimeter)

    if export_schema:
        export_path = os.path.join(output_dir, export_file)
        logger.debug(f'Writing to {export_path} ({encoding})')
        with open(export_path, 'w', encoding=encoding) as f:
            for current_type in all_types:
                f.write(current_type)
                f.write('\n')
            for current_type in all_types_reverse:
                f.write(current_type)
                f.write('\n')

    return all_types + all_types_reverse



import os
import logging
from typing import List
import pandas as pd

logger = logging.getLogger(__name__)


def generate_schema(frame: pd.DataFrame, **kwargs) -> str:
    '''
    Given the pre-processed DataFrame from the schema
    strategy, generate a schema.
    '''
    if frame is None:
        raise ValueError('frame')
    if 'column' not in frame:
        raise ValueError('column')
    if 'type' not in frame:
        raise ValueError('type')
    if 'table' not in frame:
        raise ValueError('table')
    if 'options' not in frame:
        raise ValueError('options')

    output_dir = kwargs.get('output_dir', '.')
    export_schema = kwargs.get('export_schema', False)
    export_file = kwargs.get('export_file', 'schema.txt')

    logger.debug('Concatting columns')
    frame.loc[frame['options'].isnull(), 'joined'] = frame['column'] + ': ' + frame['type'] + ' .'
    frame.loc[~frame['options'].isnull(), 'joined'] = frame['column'] + ': ' + frame['type'] + ' ' + frame['options'] + ' .'

    logger.debug('Joining Expressions into string')
    joined_expression_frame: List[str] = frame['joined'].unique().tolist()
    joined_string = '\n'.join(joined_expression_frame)

    if export_schema:
        export_file = os.path.join(output_dir, export_file)
        logger.info(f'Writing schema to {export_file}')
        with open(export_file, 'w') as f:
            f.write(joined_string)

    return joined_string



import pytest
import pandas as pd
from parameterized import parameterized
from dgraphpandas.writers.schema import generate_schema
from unittest.mock import patch, mock_open, Mock


def test_generate_schema_null_frame():
    '''
    Ensures when the frame is null, an exception is raised
    '''
    with pytest.raises(ValueError, match='frame'):
        generate_schema(None)


@parameterized.expand([
    ('without_column', pd.DataFrame(columns=['type', 'table', 'options']), 'column'),
    ('without_type', pd.DataFrame(columns=['column', 'table', 'options']), 'type'),
    ('without_table', pd.DataFrame(columns=['type', 'column', 'options']), 'table'),
    ('without_options', pd.DataFrame(columns=['type', 'table', 'column']), 'options'),
])
def test_generate_schema_bad_columns(name, frame, missing):
    '''
    Ensures the incoming dataframe does not the have the correct
    columns then an exception is raised
    '''
    with pytest.raises(ValueError, match=missing):
        generate_schema(frame)


@parameterized.expand([
    (
        'with_no_options',
        pd.DataFrame(
            columns=['column', 'type', 'table', 'options'],
            data=[
                ('id', 'string', 'customer', None),
                ('age', 'int', 'customer', None),
                ('subscription', 'uid', 'customer', None)
            ]
        ),
        '\n'.join([
            'id: string .',
            'age: int .',
            'subscription: uid .'
        ])
    ),
    (
        'with_options',
        pd.DataFrame(
            columns=['column', 'type', 'table', 'options'],
            data=[
                ('id', 'string', 'customer', '@index(exact)'),
                ('age', 'int', 'customer', None),
                ('subscription', 'uid', 'customer', '@reverse')
            ]
        ),
        '\n'.join([
            'id: string @index(exact) .',
            'age: int .',
            'subscription: uid @reverse .'
        ])
    ),
    (

        'with_multiple_types',
        pd.DataFrame(
            columns=['column', 'type', 'table', 'options'],
            data=[
                ('id', 'string', 'customer', '@index(exact)'),
                ('age', 'int', 'customer', None),
                ('subscription', 'uid', 'customer', '@reverse'),
                ('quantity', 'int', 'order', None),
                ('value', 'float', 'order', None),
                ('date', 'dateTime', 'order', '@index(day)'),
            ]
        ),
        '\n'.join([
            'id: string @index(exact) .',
            'age: int .',
            'subscription: uid @reverse .',
            'quantity: int .',
            'value: float .',
            'date: dateTime @index(day) .',
        ])
    ),
    (
        'with_duplicates_across_types',
        pd.DataFrame(
            columns=['column', 'type', 'table', 'options'],
            data=[
                ('updated_by', 'string', 'customer', None),
                ('updated_by', 'string', 'order', None),
                ('id', 'int', 'customer', '@index'),
            ]
        ),
        '\n'.join([
            'updated_by: string .',
            'id: int @index .',
        ])
    )
])
@patch('builtins.open', new_callable=mock_open)
def test_generate_schema(name, schema_frame, expected_schema, mock_file_open: Mock):
    '''
    Ensures when the schema frame is given, it matches the expected output.
    '''
    schema = generate_schema(schema_frame)
    assert schema == expected_schema
    assert not mock_file_open.called


@patch('builtins.open', new_callable=mock_open)
def test_generate_schema_export_schema(mock_file_open: Mock):
    '''
    Ensures when export schema is passed the schema is written to file
    '''
    frame = pd.DataFrame(
        columns=['column', 'type', 'table', 'options'],
        data=[
            ('id', 'string', 'customer', '@index(exact)'),
            ('age', 'int', 'customer', None),
            ('subscription', 'uid', 'customer', '@reverse'),
            ('quantity', 'int', 'order', None),
            ('value', 'float', 'order', None),
            ('date', 'dateTime', 'order', '@index(day)'),
        ]
    )

    expected_schema = '\n'.join([
        'id: string @index(exact) .',
        'age: int .',
        'subscription: uid @reverse .',
        'quantity: int .',
        'value: float .',
        'date: dateTime @index(day) .',
    ])

    schema = generate_schema(frame, export_schema=True)
    args, kwargs = mock_file_open.call_args_list[0]

    assert mock_file_open.called
    assert ('./schema.txt', 'w') == args
    assert not kwargs
    assert schema == expected_schema
